#include <immintrin.h>
#include <stdio.h>
#include <vector>
#include <chrono>
#include <iostream>
#include <cmath>

long SIZE = std::pow(2,25);

void simple_vector_example() {
        /* Initialize the two argument vectors */
    __m256 evens = _mm256_set_ps(2.0, 4.0, 6.0, 8.0, 10.0, 12.0, 14.0, 16.0);
    __m256 odds = _mm256_set_ps(1.0, 3.0, 5.0, 7.0, 9.0, 11.0, 13.0, 15.0);

    /* Compute the difference between the two vectors */
//    __m256 result = _mm256_sub_ps(evens, odds);
    __m256 result = _mm256_add_ps(evens, odds);

    /* Display the elements of the result vector */
    float* f = (float*)&result;
    printf("%f %f %f %f %f %f %f %f\n",
           f[0], f[1], f[2], f[3], f[4], f[5], f[6], f[7]);
}



void pairwise_comparison() {

    std::vector<int> vec1 = std::vector<int>(SIZE);
    std::vector<int> vec2 = std::vector<int>(SIZE);

    for (int i=0; i<SIZE; i++) {
        vec1[i] = rand() % 10000;
        vec2[i] = rand() % 10000;
    }

    auto t_start = std::chrono::high_resolution_clock::now();

    __m256i v1;
    __m256i v2;
    __m256i r1,r2;

    for (int i=0; i<SIZE; i += 8) {
        v1 = _mm256_loadu_si256((__m256i *) &vec1[i]);
        v2 = _mm256_loadu_si256((__m256i *) &vec2[i]);
        r1 = _mm256_min_epi32(v1, v2);
        r2 = _mm256_max_epi32(v1, v2);
        _mm256_storeu_si256((__m256i *) &vec1[i], r1);
        _mm256_storeu_si256((__m256i *) &vec2[i], r2);
    }

////// standard aproach
//    for (int i=0; i<SIZE; i++) {
//        if (vec1[i] > vec2[i]) {
//            int tmp = vec1[i];
//            vec1[i] = vec2[i];
//            vec2[i] = tmp;
//        }
//    }

    auto t_end = std::chrono::high_resolution_clock::now();
    double elapsed = std::chrono::duration<double, std::milli>(t_end-t_start).count()/1000.0;

    std::cout << "compared in " << elapsed << " s" << std::endl;

}

void odd_even_pair_comparison() {
    int SIZE = std::pow(2,25);
    std::vector<int> vec1 = std::vector<int>(SIZE);

    for (int i=0; i<SIZE; i++) {
        vec1[i] = rand() % 10000;
//        std::cout << vec1[i] << " ";
    }

    __m128i  mask_llhhllhh = _mm_set_epi32(0xffffffff,0,0xffffffff,0);
    __m128i  mask_hhllhhll = _mm_set_epi32(0,0xffffffff,0,0xffffffff);


    auto t_start = std::chrono::high_resolution_clock::now();

    __m128i v1;
    __m128i v2;
    __m128i r1,r2;

    for (int i=0; i<SIZE; i += 4) {
        v1 = _mm_loadu_si128((__m128i *) &vec1[i]);
        v2 = _mm_alignr_epi8(_mm_setzero_si128(), v1 ,1*4);
        r1 = _mm_min_epi32(v1, v2);
        r1 = _mm_and_si128(r1,mask_hhllhhll);
        v2 = _mm_alignr_epi8(v1, _mm_setzero_si128(),3*4);
        r2 = _mm_max_epi32(v1, v2);
        r2 = _mm_and_si128(r2,mask_llhhllhh);
        r1 = _mm_or_si128(r1,r2);
        _mm_storeu_si128((__m128i *) &vec1[i], r1);
    }

////// standard aproach
//    int tmp = 0;
//    for (int i=1; i<SIZE; i += 2) {
//        if (vec1[i-1] > vec1[i]) {
//            tmp = vec1[i-1];
//            vec1[i-1] = vec1[i];
//            vec1[i] = tmp;
//        }
//    }

    auto t_end = std::chrono::high_resolution_clock::now();
    double elapsed = std::chrono::duration<double, std::milli>(t_end-t_start).count()/1000.0;

//    std::cout << std::endl;
//    for (int i=0; i<SIZE; i++) {
//        std::cout << vec1[i] << " ";
//    }

    for (int i=1; i<SIZE; i += 2) {
        if (vec1[i-1] > vec1[i]) {
            std::cerr << "error " << std::endl;
        }
    }

    std::cout << "compared in " << elapsed << " s" << std::endl;

}



int main() {
//    simple_vector_example();
//    pairwise_comparison();
    odd_even_pair_comparison();
}